'''
Created on Nov 26, 2024

@author: demunckd
'''
import argparse
import copy
import glob
import json
import os
import sys
from cgls_cpe.config.configuration import Configuration
from cgls_cpe.logging import log
from cgls_cpe.common.helper import _str2bool
from cgls_cpe.storage.implementation.s3_api import S3PREFIX,VSIS3PREFIX
from cgls_cpe.storage import remote_storage, local_storage
from cgls_cpe.storage.implementation import s3_api
from cgls_cpe.exceptions.exception import InvalidParameterError
from cgls_cpe.cpl.date_time import getDekStart, getDekEnd, getPrevDekStart, dateNextPeriod, incrDate, today
from cgls_cpe.hspf.common_helpers import replaceChevronsInDict
from cgls_cpe.config.model.sat_image import SatImage
from cgls_cpe.config.model import enums
from cgls_cpe.config.model.area import Area
from cgls_cpe.config.model.version import Version
from cgls_cpe.config.model.timestamp import Timestamp
from magic.compat import NONE

logger = log.logger()

def parseArgs(name, description):

    parser = argparse.ArgumentParser(prog=name,
                                     description = description)
    parser.add_argument('-c', '--cfgFile', type=str, required = True, 
                        help='JSON configuration file')
    
    parser.add_argument('-p', '--productDate', type=str, required=True,
                        help='Product date in ISO format (YYYYMMDD)')
    
    parser.add_argument("--local", type=_str2bool, help="run on a local machine, or on the cluster",
                            required=False,
                            default=False)
    
    args = parser.parse_args()
    return args


def createDecadeListNum(inputStartDate, enddate, numDekadsBefore, numDekadsAfter=0, checkDate=True):
    '''createDecadeListNum
    Returns a list of dates ([first_day, enddate]) ; number of decades defined by input parameter
    
    Params:
    ------
    enddate          (string)    end period date
    numDekadsBefore  (string)    number of decades before enddate
    numDekadsAfter   (string)    number of decades after enddate
    checkDate        (bool)      perform check on startDekadeDate (default=True)
    
    Returns:
    --------
    periodLst
    
    '''
    #print("Create a list of " + str(numDekadsBefore) + " dekads up to " + str(enddate))
    periodLst = []
    # Derive the dekad start date from the end date 
    dekadStart = getDekStart(enddate)
    dekadEnd = enddate
    num = 1
    while (num <= int(numDekadsBefore)):
        if (not checkDate) or (checkDate & (int(dekadEnd) > int(inputStartDate))):
            periodLst.append(dekadEnd)
            dekadStart = getPrevDekStart(dekadEnd)
            dekadEnd = getDekEnd(dekadStart)
        num += 1

    if len(periodLst) != int(numDekadsBefore):
        print("WARNING: Period list contains " + str(len(periodLst)) +            \
                            " dekads instead of " + str(numDekadsBefore) + " expected")

    #logger.debug("Add " + str(numDekadsAfter) + " dekads after " + str(enddate))
    if int(numDekadsAfter) > 0:
        print("Add " + str(numDekadsAfter) + " dekads after " + str(enddate))
        num = 1
        dekadEnd = enddate
        while( num <= int(numDekadsAfter)):
            dekadStart = dateNextPeriod(dekadEnd, 10)
            dekadEnd   = getDekEnd(dekadStart)
            periodLst.append(dekadEnd)
            num += 1

    return sorted(periodLst, reverse=False)

def createConsoDates(cfgDict, dekad, nb_dekads=7):
    '''createConsoDates
    Returns a list of dates for the period2Process and the associated consolidations
        only maintain the consolidations to be processed
    
    Params:
    ------
    cfgDict      (dict)    holding configuration parameters
    dekad        (string)  the dekad date to process
    nb_dekads    (int)
    
    Returns:
    --------
    _periods2Process, _consoDateDic
    '''
    _consoDateDic   = {}
    _periods2Process = createDecadeListNum(cfgDict["inputStartDate"], dekad, nb_dekads)
    # Create dict for RT0, RT1, .. RT6
    #rt0Date = _periods2Process[-1]
    for idx, date in enumerate(reversed(_periods2Process)):
        if 'RT'+str(idx) in cfgDict["consos"]:
            _consoDateDic['RT'+str(idx)]   = date
    return _periods2Process, _consoDateDic

def _dateToISOdate(date):
    """ Turns a YYYYMMDD date into YYYY-MM-DD
    
    Parameters
    ----------
    date : str
    
    Returns
    -------
    str:
        date in ISO format YYYY-MM-DD
    """
    return f'{date[0:4]}-{date[4:6]}-{date[6:8]}'

def createPostCfgDict(cfgDict, roiInfoOut, roiCollInfoIn, logger, conso=None, consoDate=None):
    """Create a mosaic dictionary 
    
    Substitution variables (< >) will also be resolved.
    
    Parameters
    ----------
    cfgDict : dict
        configuration dictionary which should contain the following keys:
        * 'productDate': (str) product reference data in ISO format YYYYMMDD
        * 'periodicity': (str) periodicity of the product
        * 'inFolder': (str) path to search in, including substitution variable
        * 'inFilePattern': (str) file pattern to search for, including substitution variable
        * 'missingTileFolder': (str) path to search missing tiles, including substitution variable. Set to false to ignore missing tiles
,       * 'missingTilePattern': (str) missing file pattern to search for, including substitution variable. Set to false to ignore missing tiles
        * 'outFolder': (str) path to store output mosaic file, including substitution variable
        * 'outFilePattern': (str) output file pattern, including substitution variable
        * 'outputRoiFile': (str) full path (folder & filename) to GeoJSON roi file specifying output ROI's
        * 'outputRoiId': (list) the ROI id to be used from the outputRoiFile
        * 'overWriteExisting': (bool) true -overwrite existing output files, false - skip existing output files
        * 'settings': (dict) netCDF generation settings
            ** 'fileAttributes': (dict) Key value pairs to be added as metadata on file level 
            ** 'bandList' (list of dict) each dictionary contains specific information (band mapping, metadata) for a band
    roi : str
        The output ROI name
    roiCollInfoIn : dict
        The input ROI collection information
    logger : object
        Instance to log to
    conso : str
        The consolidation to process, defaults to None
    
    Returns
    -------
    dict:
        dictionary containing the following
            "input_files"          : netcdf input files
            "output_file"          :
            "output_roi_coords"    :
            "bands"                :
            "output_band_order"    :
            "output_pixel_step"    :
            "netcdf_time_ref_date" :
            "attributes"           :
            "productDate"          :
            "Optional"             : ''
            "input_band_order"     :

    """
    logger(f" > Creating mosaic configurations for {roiInfoOut['id']} ")
    startDateISO = incrDate(cfgDict['productDate'], days=-cfgDict['inputDaysBefore'], outFormat='%Y-%m-%d')
    endDateISO = incrDate(cfgDict['productDate'], days=-cfgDict['inputDaysAfter'], outFormat='%Y-%m-%d')
    replacementDict = {
            '<roi>': roiInfoOut['id'],
            '<roiCollection>': roiInfoOut['id'].lower(),
            '<startDateISO>': startDateISO,
            '<endDateISO>': endDateISO,
            }
    if 'timeliness' in cfgDict:
        replacementDict['<timeliness>'] = cfgDict['timeliness']
    if 'subtypes' in cfgDict:
        replacementDict['<subtypes>'] = cfgDict['subtypes']
    productDate = cfgDict['productDate']
    if conso:
        replacementDict['<consoDate>'] = consoDate
        replacementDict['<consoDateISO>'] = _dateToISOdate(consoDate)
        replacementDict['<consoYear>'] = consoDate[0:4]
        replacementDict['<consoMonth>'] = consoDate[4:6]
        replacementDict['<consoDay>'] = consoDate[6:8]
        replacementDict['<conso>'] = conso
        productDate = consoDate
    mosaicDict = replaceChevronsInDict(copy.deepcopy(cfgDict), replacementDict)
    
    timeliness = None
    if 'timeliness' in cfgDict:
        timeliness = enums.Timeliness(cfgDict['timeliness'])
    subtypes = None
    if 'subtypes' in cfgDict:
        subtypes = enums.Subtype(cfgDict['subtypes'])
    consolidation = None
    if conso:
        consolidation = enums.Consolidation(conso)
    year    = productDate[:4]
    month   = productDate[4:6] 
    day     = productDate[6:8]
    timestamp = Timestamp(int(year), int(month), int(day), int(0), int(0), int(0))
    
    inputProduct = SatImage(prodType=enums.ProductType(cfgDict['productType']), 
                            subtypes=subtypes,
                            periodicity=getPeriodicity(cfgDict['periodicity']), 
                            timestamp=timestamp,
                            area=Area('<tile>'), 
                            sensor=enums.SensorAbbr(cfgDict['sensor']),
                            version= Version.from_string(cfgDict['inputVersion']),
                            conso=consolidation,
                            timeliness=timeliness,
                            ext=enums.FileTypeExtension(cfgDict['inFileExt'])
                            )
    inputLocation = remote_storage.get_full_URL(inputProduct)
    
    #inputFiles, lastInputFile= discoverInputs(mosaicDict, roiCollInfoIn, productDate, inputLocation, logger)
    inputFiles,_ = discoverInputs(mosaicDict, roiCollInfoIn, productDate, inputLocation, logger)
    if len(inputFiles) == 0: #check becomes obsolete due to check in discoverInputs
        logger(f"     > No input files found")
        return None
    '''
    outputProduct = SatImage.create_from_filename(os.path.basename(lastInputFile), 
                                             area=Area(roiInfoOut['id']),
                                             periodicity=getPeriodicity(cfgDict['periodicity']),
                                             version=Version.from_string(cfgDict['outputVersion']),
                                             is_dissemination=True)
    '''
    outputProduct = SatImage(prodType=enums.ProductType(cfgDict['productType']), 
                            subtypes=subtypes,
                            periodicity=getPeriodicity(cfgDict['periodicity']), 
                            timestamp=timestamp,
                            area=Area(roiInfoOut['id']), 
                            sensor=enums.SensorAbbr(cfgDict['sensor']),
                            version= Version.from_string(cfgDict['outputVersion']),
                            conso=consolidation,
                            timeliness=timeliness,
                            ext=enums.FileTypeExtension(cfgDict['outFileExt']),
                            is_dissemination=True
                            )
    outputFile = remote_storage.get_full_URL(outputProduct)
    
    mosaicDict.update({
            "input_files"          : inputFiles,
            "output_file"          : outputFile,
            "output_roi_coords"    : roiInfoOut.geometry.coordinates,
            "bands"                : mosaicDict['settings']['bandList'],
            "output_band_order"    : list(mosaicDict['settings']['bandList'].keys()),
            "output_pixel_step"    : 336,
            "netcdf_time_ref_date" : "19700101",
            "attributes"           : mosaicDict['settings']['fileAttributes'],
            "productDate"          : mosaicDict['productDate'],
    })
    
    return mosaicDict

def createPostCfgDictList(cfgDict, logger):
    """Create a list of mosaic configuration dictionaries to spark on the cluster
    
    Substitution variables (< >) will also be resolved.
    
    Parameters
    ----------
    cfgDict : dict
        configuration dictionary which should contain the following keys:
        * 'productDate': (str) product reference data in ISO format YYYYMMDD
        * 'inFolder': (str) path to search in, including substitution variable
        * 'inFilePattern': (str) file pattern to search for, including substitution variable
        * 'missingTileFolder': (str) path to search missing tiles, including substitution variable. Set to false to ignore missing tiles
,       * 'missingTilePattern': (str) missing file pattern to search for, including substitution variable. Set to false to ignore missing tiles
        * 'outFolder': (str) path to store output mosaic file, including substitution variable
        * 'outFilePattern': (str) output file pattern, including substitution variable
        * 'outputRoiFile': (str) full path (folder & filename) to GeoJSON roi file specifying output ROI's
        * 'outputRoiId': (list) the ROI id to be used from the outputRoiFile
        * 'overWriteExisting': (bool) true -overwrite existing output files, false - skip existing output files
        * 'settings': (dict) netCDF generation settings
            ** 'fileAttributes': (dict) Key value pairs to be added as metadata on file level 
            ** 'bandList' (list of dict) each dictionary contains specific information (band mapping, metadata) for a band
    logger : object
        Instance to log to
    
    Returns
    -------
    list of dict:
        A list of mosaic configuration dictionaries
    """
    import geojson
    logger("Creating configurations for mosaics")    
    # now create the mosaic configurations:
    mosaicCfgDictList = []
    # Get the ROI info
    logger(f" > Determine output ROI")
    with open(cfgDict['outputRoiFile']) as roiFile:
        roiCollInfoOut = geojson.loads(roiFile.read())
    with open(cfgDict['inputRoiFile']) as roiFile:
        roiCollInfoIn = geojson.loads(roiFile.read())
    for feature in roiCollInfoOut.features:
        if feature["id"] in cfgDict['outputRoiId']:
            roiInfo = feature
            logger(f"   > Output ROI {feature['id']} used")
            # now create the mosaic configurations:
            if 'consos' in cfgDict and len(cfgDict['consos']) > 0:
                logger(f" > Creating mosaic configurations for {cfgDict['consos']}")
                _, consoDateDic= createConsoDates(cfgDict, cfgDict['productDate'], 7)
                for conso in cfgDict['consos']:
                    mosaicCfgDict = createPostCfgDict(cfgDict, roiInfo, roiCollInfoIn, logger, conso, consoDateDic[conso])
                    if mosaicCfgDict:
                        mosaicCfgDictList.append(mosaicCfgDict)
            else:
                logger(f" > Creating mosaic configurations")
                mosaicCfgDict = createPostCfgDict(cfgDict, roiInfo, roiCollInfoIn, logger)
                if mosaicCfgDict:
                    mosaicCfgDictList.append(mosaicCfgDict)
    
    return mosaicCfgDictList

def setupApplication(args,skip_output_replace=False):
    
    # Create the configuration dictionary based on the configuration file and the date input
    if ( s3_api.is_s3(args.cfgFile)):
        cfgDict = json.loads(remote_storage.get_contents_as_string(args.cfgFile))
    else:
        with open(args.cfgFile, 'r') as cfgFile:
            cfgDict = json.load(cfgFile)
    
    if sys.path[0] != getBasePythonPath() :
            sys.path.insert(0,getBasePythonPath())
    
    cfgDict['productDate'] = args.productDate
    year = cfgDict['productDate'][0:4]
    month = cfgDict['productDate'][4:6]
    day = cfgDict['productDate'][6:8]
    processDateISO = _dateToISOdate(today())
    replacementDict = {
            '<env>': Configuration().get_environment_name(),
            '<sensor>': cfgDict['sensor'],
            '<producttype>': cfgDict['productType'],
            '<producttype_lc>': cfgDict['productType'].lower(),
            '<inputversion>': cfgDict['inputVersion'],
            '<outputversion>': cfgDict['outputVersion'],
            '<productDate>': cfgDict['productDate'],
            '<year>': year,
            '<month>': month,
            '<day>': day,
            '<productDateISO>': _dateToISOdate(cfgDict['productDate']),
            '<processDateISO>': processDateISO,
            '<processYear>': processDateISO[0:4],
            '<periodicity>': str(cfgDict['periodicity'])
            }
        
    cfgDict = replaceChevronsInDict(cfgDict, replacementDict)
    
    return cfgDict

def getBasePythonPath():
        return os.path.dirname(__file__) + '/workflow/utils'

def getPeriodicity(periodicity):
    if periodicity == 1:
        return enums.Periodicity.DAILY
    elif periodicity == 10:
        return enums.Periodicity.DEKAD
    elif periodicity == 30:
        return enums.Periodicity.MONTHLY
    else:  #cfgDict['periodicity'] == 365:
        return enums.Periodicity.YEARLY

def discoverInputs(cfgDict, roiCollInfoIn, productDate, inputLocation, logger=print):
    """Discover input products

    The function will search for all the input products for given productDate
    The following substitution variables will be resolved when searching for files: 

    Parameters
    ----------
    cfgDict : dict
        configuration dictionary which should contain the following keys:
        * 'productDate' : (str) product date ISO format YYYYMMDD
        * 'inFolder': (str) path to search in, including substitution variable
        * 'inFilePattern': (str) file pattern to search for, string is fully resolved
    logger : object
        Instance to log to, defaults to print

    Returns
    -------
    list of str:
        A list of all the paths to the input files
    """
    logger(f"Discovering inputs for {productDate}")
    remoteFileLst = []
    nrTilesFound = 0
    lastInputFile = '' #keep track in order to have a means of defining output filename
    for roiInfo in roiCollInfoIn.features:
        filePattern = inputLocation.replace('<tile>', roiInfo['id'])
        #logger("We are going to sync to local storage")
        #remote_folder = filePattern[0:filePattern.rindex('/')]
        #logger("Syncing from " + remote_folder + " to " + tmp_input)
        #rslt = remote_storage.sync_remote_to_local_folder(tmp_input,remote_folder) #returns all tiles for all consolidations - wildcard not possible
        rslt = remote_storage.list_objects(filePattern)
        if len(rslt) > 0: #check if available
            remoteFileLst.extend(rslt)
            nrTilesFound += 1
            lastInputFile = rslt[0]
        else:
            # get missing tile
            if len(cfgDict['missingTileFolder']) > 0:
                missingFilePattern = os.path.join(cfgDict['missingTileFolder'], cfgDict['missingTilePattern'].replace('<tile>', roiInfo['id']))
                #path should be OK, just make sure it is available in ObjectStore
                rslt = remote_storage.list_objects(missingFilePattern)
                if len(rslt) > 0:
                    remoteFileLst.extend(rslt)
    logger(f" > Discovered {len(remoteFileLst)} inputs of which {nrTilesFound} tiles and {len(remoteFileLst) - nrTilesFound} missing tiles")
    if nrTilesFound < cfgDict['minNumberInputTiles']:
        logger(f"Did not found enough input products {nrTilesFound}/{cfgDict['minNumberInputTiles']}, leaving it here")
        raise InvalidParameterError("Not enough products found! Check your config")
    
    return remoteFileLst, lastInputFile

def syncResultsToRemote(cfgDict,logger=print):
    '''
        srcFolder = cfgDict['outFolder'].replace('<tile>','')
        tgtFolder = cfgDict['outFolderRemote'].replace('<tile>','')
        logger('Syncing from ' + srcFolder + ' to ' + tgtFolder)
        result = remote_storage.sync_local_folder_to_remote(srcFolder, tgtFolder)
        if result != 0:
            raise Exception("Could not upload result!")
    '''
    srcFile = cfgDict['output_file']
    tgtFile = os.path.join(cfgDict['out_folder_remote'], cfgDict['out_filename'])
    rslt = remote_storage.upload_single_file_to_remote_path(srcFile, tgtFile)
    if rslt != 0: #rslt always 0??
        raise Exception("Could not upload result!")

