'''
Created on Dec 21, 2023

@author: demunckd
'''
from cgls_cpe.common import helper
import shapely
from shapely.geometry import Polygon, MultiPolygon, LineString

from cgls_cpe.logging import log

logger=log.logger()

import os.path
import re

def convert_sen_tile_to_prob(tile:str):
    first_part = tile[0:3]
    probav_y_index = int(tile[4:6])-1
    if probav_y_index < 0:
        return None
    second_part = 'Y' +  helper.pad_two_digits(probav_y_index )
    result = first_part + second_part 
    return result

def convert_sen_tile_list_to_prob(list_of_sen_tiles:str):
    new_list = []
    for tile in list_of_sen_tiles:
        new_tile =convert_sen_tile_to_prob(tile)
        if new_tile is not None:
            new_list.append(new_tile)
    return new_list             


def convert_prob_tile_to_sen(tile:str):
    first_part = tile[0:3]
    sen_y_index = int(tile[4:6])+1
    second_part = 'Y' +  helper.pad_two_digits(sen_y_index )
    result = first_part + second_part 
    return result


def get_all_tiles_sen():
        list_of_tiles = []
        for x_index in range(36):
            for y_index in range(15):
                tile = 'X' + helper.pad_two_digits(x_index) +\
                        'Y' + helper.pad_two_digits(y_index)
                list_of_tiles.append(tile )

        return list_of_tiles
    
def get_s3GeoJsonFullGrid():
    return os.path.join( os.path.dirname(__file__),'cgl_s3_grid-geo.json' )

def get_pbvGeoJsonFullGrid():
    return os.path.join( os.path.dirname(__file__),'cgl_pv_grid_333m-geo.json' )

def showTiles(title, json_file, list_of_colors, list_of_list_of_tiles):
    #we only want to import these when really needed
    import matplotlib
    matplotlib.use('TkAgg')
    import matplotlib.pyplot as plt
    import geopandas as gpd


    
    input_grid_json = gpd.read_file(json_file)
    #From GeoPandas, our world map data
    try:
        worldmap = gpd.read_file(gpd.datasets.get_path("naturalearth_lowres"))
    except:
        #for this to work, you need to do a conda install geodatasets
        import geodatasets
        dataset_key = geodatasets.data.flatten().keys()
        path=geodatasets.get_path('naturalearth.land')
        logger.debug("reading path: " + path)
        worldmap = gpd.read_file(path)
        logger.debug("done")

    # Creating axes and plotting world map
    fig, ax = plt.subplots(figsize=(20, 10))
    worldmap.plot(color="lightgrey", ax=ax)

    # Creating axis limits and title
    plt.xlim([-180, 180])
    plt.ylim([-90, 90])

    plt.title(title)
    plt.xlabel("Longitude")
    plt.ylabel("Latitude")

    i = 0
    while i < len(list_of_colors):
        tiles_in_grid = input_grid_json[input_grid_json.id.isin(list_of_list_of_tiles[i])]
        tiles_in_grid.plot(ax=ax, color=list_of_colors[i], alpha=0.3, edgecolor="black", label='')
        i+=1

    plt.show()


class Geometry():
    """
    This class can be used to calculate bbox, split the geometry over the dateline
    """

    def __init__(self):
        self.geometry = None
        self.xmin = None
        self.xmax = None
        self.ymin = None
        self.ymax = None

        #global boundaries
        self.maxLat = 85
        self.minLat = -85
        max_coords = ((-180, self.maxLat), (-180, self.minLat), (180, self.minLat), (180, self.maxLat), (-180, self.maxLat))
        self.glo_bbox = Polygon(max_coords)

    def footprint_2_polygon(self, footprint, topping=True):
        '''Convert footprint string into a multipolygon object
        if topping is enabled, we will top the image at +/- 85 degrees
        '''

        polygons = []   #this can be a multipolygon or a single one
        #pattern = "\(([0-9\,\.\s\-]*)\)"
        pattern = r"\(([0-9,.\s\-]*)\)"

        footprints = re.findall(pattern, footprint)     #find all polygons, 1 for a simple Polygon object and multiple for a MultiPolygon object
        logger.debug('footprints %s' % len(footprints))
        polygon_lst = []
        for footprint in footprints:
            coords_str = footprint
            coords_pairs = coords_str.split(',')        #split the string into a list of x,y coordinates

            # Create a list of tuples (longitude, latitude) by converting the strings to floats
            coords = [(float(coord.split()[0]), float(coord.split()[1])) for coord in coords_pairs]

            #make it a shapely object and get rid of the north/south areas if needed

            polygon = Polygon(coords)
            #to fix an artifact in the geometry, skip polygons having an area <1 and a lat of the centroid < 84
            if (polygon.area < 1) and ( (polygon.centroid.y < -84) or (polygon.centroid.y > 84)):
                logger.debug('skipping polygon having the following geometry : %s' % shapely.to_geojson(polygon))
                continue
            #polygon_split = self.split_polygon_by_dateline(polygon)
            polygon_split = self.split_on_date_line(polygon)        #this is always a multipolygon
            for single_polygon in list(polygon_split.geoms):
                logger.debug('single polygon : %s' % shapely.to_geojson(single_polygon))
                polygon_topped = single_polygon.intersection(self.glo_bbox)   #this can generate geometrycollections  #split over dateline???
                if polygon_topped.geom_type =='LineString':
                    print("LineString was detected. Skipping...")
                    pass
                elif polygon_topped.geom_type =='GeometryCollection' or polygon_topped.geom_type == 'MultiPolygon':
                    for polygon in list(polygon_topped.geoms):
                        if polygon.geom_type == 'Polygon':
                            #skipping LineStrings here, it can be introduced by the shapely.intersection method
                            polygon_lst.append(polygon)

                else:
                    polygon_lst.append(polygon_topped)

        # Create the Shapely Polygon
        logger.debug(polygons)
        self.geometry = MultiPolygon(polygon_lst)
        return self.geometry

    def split_on_date_line(self, my_polygon):
        '''
        The incoming polygon will be cut at the dateline. So a frame passing the dateline will be cut into a positive and a negative part.
        In a nutshell it works like this:
        1) move coordinates passing the dateline towards the dateline
        2) clean the coordinates at the dateline that are no longer needed(only keep the first and last of that dateline serie)
        :param my_polygon:
        :return: MultiPolygon
        '''

        pos_coords = []
        neg_coords = []

        if my_polygon.geom_type == 'MultiPolygon':
            #do not split a multipolygon
            multipolygon = my_polygon

        else:
            logger.debug("before splitting: %s " % shapely.to_geojson(my_polygon))

            date_crossing = self.crosses_dateline(my_polygon)   #count of the frame is passing the dateline
            polygon_topped = self.pole_topping(my_polygon)      #cut off the highest/lowest latitudes

            if date_crossing > 1:
                # date_line crossing around equator results in 2 overpasses
                coordinates_topped = list(polygon_topped.exterior.coords)

                #move all coordinates passing the dateline directly at the dateline
                for coord in coordinates_topped:
                    if coord[0] < 0 :
                        pos_coords.append((180, coord[1]))
                        neg_coords.append((coord[0], coord[1]))
                    else:
                        pos_coords.append(( coord[0] , coord[1]))
                        neg_coords.append((-180, coord[1]))


                # cleanup coordinates(that we just have moved)on the dateline. Just use the first and last of the series.
                polygon_east = None
                polygon_west = None
                try:
                    polygon_east = self.cleanup_coordinates(pos_coords, 180)
                except:
                    logger.warn("Could not cleanup east coordinate: %s" , pos_coords)
                try:
                    polygon_west = self.cleanup_coordinates(neg_coords, -180)
                except:
                    logger.warn("Could not cleanup west coordinate: %s" , neg_coords)                    
                
                if polygon_east is not None and polygon_west is not None:    
                    multipolygon = MultiPolygon([polygon_east, polygon_west])
                elif polygon_east is not None: 
                    multipolygon = MultiPolygon([polygon_east])
                elif polygon_west is not None:
                    multipolygon = MultiPolygon([polygon_west])
                else:
                    multipolygon = MultiPolygon([polygon_topped])

            else:
                polygon_frame1_shapely = polygon_topped
                polygon_frame2 = None
                multipolygon = MultiPolygon([polygon_frame1_shapely])

        return multipolygon

    def cleanup_coordinates(self, coords, lon_2_remove):
        '''
        This function will cleanup coordinates upon the dateline. Only the first and last of the list will remain.
        :param coords: All coordinates as a list of tuples
        :param lon_2_remove: 180 or -180, the coordinates having this value will be reduced
        :return: polygon
        '''
        logger.debug("cleanup coordinatew with longitude=%s" % lon_2_remove)
        date_line_indexes = [index for index, (lon, lat) in enumerate(coords) if lon == lon_2_remove]
        keep_indexes = [index for index, (lon, lat) in enumerate(coords) if lon != lon_2_remove]

        # We have 2 cases here,
        # 1) the keep coordinates are surrounded by dateline coordinates
        # D D K K D D (K=Keep, D=dateline) => D K K D
        if len(keep_indexes) == keep_indexes[-1] - keep_indexes[0]+1:
            first_keep_index = min(keep_indexes)
            last_keep_index = max(keep_indexes)
            reduced_coordinates = [coords[x] for x in keep_indexes]  # get coordinates not on dateline
            reduced_coordinates = [coords[first_keep_index - 1]] + reduced_coordinates + [
                coords[last_keep_index + 1]] + [coords[first_keep_index - 1]]
            polygon = Polygon(reduced_coordinates)


        else:
            # 2) the dateline coordinates are surrounded by Keep values
            # K K D D K K
            first_dateline_index = min(date_line_indexes)
            last_dateline_index = max(date_line_indexes)

            index_2_remove = first_dateline_index + 1
            nr_of_dels = last_dateline_index - first_dateline_index -1
            for index in range(nr_of_dels):
                del coords[index_2_remove]  #it is always the same index to remove as elements are shifting with easht removal
            polygon = Polygon(coords)
        return polygon

    def crosses_dateline(self, polygon):
        # a frame passing the equator has 2 decent passes
        # a frame passing the poles, only has 1 decent pass around the dateline,is it? Sound strange now...
        # Get the coordinates of the polygon's exterior ring
        coords = list(polygon.exterior.coords)
        nr_passes = 0
        # Iterate through the coordinates
        for i in range(len(coords) - 1):
            lon1, lat1 = coords[i]
            lon2, lat2 = coords[i + 1]

            # Check if the difference in longitude crosses the dateline
            if abs(lon1 - lon2) > 180:
                nr_passes += 1

        return nr_passes

    def pole_topping(self, polygon):
        '''
        Cut of the incoming polygon the the min and max latitude defined the Geometry class
        :param polygon:
        :return:a tuple of coordinates that have been topped to the max/min latitudes
        '''
        coords = list(polygon.exterior.coords)
        pos_coords = []
        for i in range(len(coords)):
            lon1, lat1 = coords[i]
            if lat1 > self.maxLat:
                lat1 = self.maxLat
            elif lat1 < self.minLat:    #probably not needs as I have found a frame below this value
                lat1 = self.minLat
            pos_coords.append((lon1, lat1))

        polygon = Polygon(pos_coords)
        return polygon


    #
    # def split_polygon_by_dateline(self, polygon):
    #     # Normalize longitudes to be within -180 to 180 range
    #     normalized_coords = []
    #     for x, y in polygon.exterior.coords:
    #         if x > 180:
    #             x = x - 360
    #         normalized_coords.append((x, y))
    #
    #     normalized_polygon = Polygon(normalized_coords)
    #
    #     # Define the Date Line as a vertical line at 180° longitude
    #     dateline = LineString([(180, -90), (180, 90)])
    #
    #     # Split the polygon by the Date Line
    #     split_polygons = split(normalized_polygon, dateline)
    #
    #     # Re-normalize the longitudes of the split polygons
    #     result_polygons = []
    #     for part in split_polygons:
    #         renormalized_coords = []
    #         for x, y in part.exterior.coords:
    #             if x < -180:
    #                 x = x + 360
    #             renormalized_coords.append((x, y))
    #         result_polygons.append(Polygon(renormalized_coords))
    #
    #     return MultiPolygon(result_polygons)
    #
