#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Copernicus Global Land - mosaic .

:author: frederik.coelst@vito.be>

Notes
-----
Documented with Sphinx in numpy style:
 #. https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_numpy.html
 #. https://numpydoc.readthedocs.io/en/latest/format.html#docstring-standard

History
-------
========  =======  ======  ========  ========
date      version  author  jira ref  Comments
========  =======  ======  ========  ========
20200831  1.0.0    CoF     GIOG-979  S3 POST: Implement mosaic
20201005  1.0.1    RVdG    GIOG-1366 S3 GEO: update metadata
                                        update history, add history of input product
20201013  1.0.2    RVdG    GIOG-1416 S3 GEO: update metadata (2)
                                        update time dimension
                                        update history 
                                        fix for fillValue==0
20201029  1.0.3    RVdG    GIOG-1443 S3 POST: OLCI products have incorrect time coordinate value
                                        remove +1 when calculating daysSinceRef in method createOutputNc
20201110  1.0.4    RVdG    GIOG-1435 S3 POST: mosaic output to be packed
                                        add complevel when calling createVariable
20201214  1.0.5    RVdG    GIOG-1282 S3 POST: Setup continental upload to EUMETCast
                                        quickfix instead of implementing intersect
                                        writeBandNc: if inStartCol,inStartRow,inEndCol,inEndRow are not consistent, do nothing
20210104  1.0.6    RVdG    GIOG-1524 NDVI 1KM v3: Post processing, add band rename option
20210104  1.0.7    RVdG    GIOG-1524 NDVI 1KM v3: Post processing, add band rename option
                                        review comment
20210426  1.0.8    RVdG    GIOG-1850 S3 POST: error mosaic when band attribute units=days
                                        make use of attribute decode_time=True for time related bands, decode_time=False for others
20220627  1.0.9    RVdG    GIOG-3144 remove crs:GeoTransfor attribute
20241206  1.1.0    RVdG    GIOG-3741 Cloudify global mosaic input netCDF, output netCDF
========  =======  ======  ========  ========

"""
__version__ = '1.1.0'
__date__ = '2024-12-06'

# Regular imports
import os, sys
import numpy   as np
import netCDF4 as nc
import xarray  as xr
import argparse
import json
# Vito imports
from cgls_cpe.cpl.netcdf import createLatLonNetCDF, splitInOutOfBounds

def convertRoiToLatLon(polygonCoords,resolution):
    #get corner values
    west  = polygonCoords[0][0]
    north = polygonCoords[0][1]
    south = polygonCoords[1][1]
    east  = polygonCoords[3][0]
    longitudes = np.arange(start=west, stop=east, step = resolution, dtype = np.float64)
    latitudes  = np.arange(start=north,stop=south,step = -resolution,dtype = np.float64)
    return longitudes,latitudes

def getNetCDF4Dtype(filename,band):
    with nc.Dataset(filename,'r') as dataset:
        return dataset.variables[band].dtype

def xrOpenDatasetsAndRead(filename, inVars2readLst, offset=None, shape=None, mask=None, decode_time=False):
    netCdf4VarDict = {}
    #don't apply auto mask & scale!!!
    dataset = xr.open_dataset(filename, mask_and_scale=False, decode_times=decode_time)
    for var in inVars2readLst:
        if None not in [offset,shape]:
            chunkXOff, chunkYOff  = offset
            chunkXSize,chunkYSize = shape
            chunkXEnd = chunkXOff + chunkXSize
            chunkYEnd = chunkYOff + chunkYSize
            bandDS    = dataset.get(var)[chunkXOff:chunkXEnd,chunkYOff:chunkYEnd]
        elif type(mask) is np.ndarray:
            #TODO: try to apply the mask to the xarray dataArray
            # *bandDS=dataset.get(var)[np.where(mask)]
            #  bandDS.shape
            #  (2524090, 2524090)  -->(eens zo groot & 2D + memory error bij uitlezen?!!!)
            # *bandDS=dataset.get(var).where(xr.DataArray(mask,dims=dataset.get(var).dims),drop=True)
            #  bandDS.shape
            #  (5247, 512)
            #  5247*512=2686464    -->(te groot???)
            # NetCDF4 does it with dataset.variables[var][np.where(segChunkMask)]
            # but is way too slow and has memoryerror with chunk>=5040
            bandDS = xr.DataArray(dataset.get(var).values[mask])
        else:
            bandDS = dataset.get(var)
        #PATCH: xarray doesn't detect dtype correctly
        varDtype = getNetCDF4Dtype(filename,var)
        netCdf4VarDict[var] = bandDS.values.astype(varDtype)
    dataset.close()
    return netCdf4VarDict

def readNetCDF4(filename,inVars2readLst,offset=None,shape=None,mask=None):
    netCdf4VarDict = {}
    
    #GIOG-1850
    #handle 'time' with decode_times=True
    if 'time' in inVars2readLst:
        inVars2readLst.remove('time')
        netCdf4VarDict.update(xrOpenDatasetsAndRead(filename, ['time'], offset, shape, mask, decode_time=True))
        
    #handle the rest of inVars2readLst with decode_times=False
    netCdf4VarDict.update(xrOpenDatasetsAndRead(filename, inVars2readLst, offset, shape, mask, decode_time=False))
    
    return netCdf4VarDict

def createMosaicNetcdf(mosaicDict, logger=print):
    """
    createMosaicNetcdf creates a mosaic netCDF output file based on the input dictionary
    Parameters
    ----------
    mosaicDict : dict
        dictionary containing the following
            "input_files"          : netcdf input files
            "output_file"          :
            "output_roi_coords"    :
            "bands"                :
            "output_band_order"    :
            "output_pixel_step"    :
            "netcdf_time_ref_date" :
            "attributes"           :
            "productDate"          :
            "Optional"             : ''
            "input_band_order"     :
    """
    logger('Creating mosaic netCDF')
    #get the values from the dictionary
    inputFileLst       = mosaicDict['input_files'         ]
    outputFile         = mosaicDict['output_file'         ]
    outputRoiCoords    = mosaicDict['output_roi_coords'   ]
    outputBandsDict    = mosaicDict['bands'               ]
    outputAttributes   = mosaicDict['attributes'          ]
    if 'input_band_order' in mosaicDict: #make backward compatible
        inputBandLst   = mosaicDict['input_band_order'    ]
    else:
        inputBandLst   = mosaicDict['output_band_order'   ]
    outputBandLst      = mosaicDict['output_band_order'   ]
    outputPixelStep    = mosaicDict['output_pixel_step'   ]
    outputRefDate      = mosaicDict['netcdf_time_ref_date']
    outputProdDate     = mosaicDict['productDate'         ]
    
    # Check for outputPath folder
    outputPath         = os.path.dirname(outputFile)
    if not os.path.isdir(outputPath):
        os.makedirs(outputPath, exist_ok=True)
    tmpFile = outputFile.replace('.nc','_tmp.nc')
    if os.path.exists(outputFile):
        os.remove(outputFile)
    # calculate some values
    # get the shape (based on resolution)
    outputResolution   = 1 / outputPixelStep
    outputChunking     = None
    # filter attributes
    globalAttrKeys     = outputAttributes.keys() - outputBandLst
    outputAttrGlobal   = {k:outputAttributes[k] for k in globalAttrKeys}
    #add name and CLASS key & convert to list
    [outputBandsDict[b].update({'name':b}) for b in outputBandLst]
    outputBandsInfoLst = [outputBandsDict[b] for b in outputBandLst]
    outputLatLonCoords = convertRoiToLatLon(outputRoiCoords[0],outputResolution)
    outputLongitudes   = outputLatLonCoords[0]
    outputLatitudes    = outputLatLonCoords[1]
    outputNorthLat     = outputRoiCoords[0][0][1]
    outputWestLon      = outputRoiCoords[0][0][0]
    
    #create the mosaic file
    createLatLonNetCDF(tmpFile, outputLongitudes, outputLatitudes,
                       outputBandsInfoLst, globalAttributes = outputAttrGlobal,
                       chunking=outputChunking,
                       refDate=outputRefDate, prodDate=outputProdDate)
    # add tiles to mosaic
    with nc.Dataset(tmpFile, 'a') as outFile:
        outFile.set_auto_maskandscale(False) #Otherwise, data is scaled and offset before putting in the dsout
        #read each input file & write into mosaic
        for i, inputFile in enumerate(inputFileLst):
            inputBands     = ['lat','lon'] + inputBandLst
            inputArrDict   = readNetCDF4(inputFile,inputBands)
            bandLat        = inputArrDict.pop('lat')
            bandLon        = inputArrDict.pop('lon')
            #check for 'time' dimension & add 1 if needed
            if 'time' not in inputArrDict:
                inputArrDict = {k:np.expand_dims(v,axis=0)      \
                                for k,v in inputArrDict.items() }
            _,lines,columns = inputArrDict[outputBandLst[0]].shape
            #TODO: nearset neighbour integer?
            lineOffset          = (outputNorthLat - bandLat.max())/outputResolution
            lineOffset          = np.rint(lineOffset).astype(np.int64)
            columnOffset        = (bandLon.min()  - outputWestLon)/outputResolution
            columnOffset        = np.rint(columnOffset).astype(np.int64)
            lineEnd             = lineOffset   + lines
            columnEnd           = columnOffset + columns
            for bandId, outputBand in enumerate(outputBandLst):
                inputBand      = inputBandLst[bandId]
                inputBandArray = inputArrDict.pop(inputBand)
                #select output dataset
                dsOut          = outFile.variables[outputBand]
                inputShape     = inputBandArray.shape[-2:]
                outputShape    = dsOut.shape[-2:]
                #get the indices that fit inside our ROI
                #note: coordinates here are [xmin,ymin,xmax,ymax]
                imgSizeXY      = (outputShape[1],outputShape[0])
                inputIdxLst    = [0,0] + list(inputShape)
                outputIdxLst   = [columnOffset,lineOffset,columnEnd,lineEnd]
                outputInsideIdxLst,invalidIdxLst = splitInOutOfBounds(outputIdxLst,  \
                                                                          imgSizeXY      )
                outStartCol,outStartRow,outEndCol,outEndRow = outputInsideIdxLst
                inStartCol,inStartRow,inEndCol,inEndRow     = [(v - invalidIdxLst[i]) for i,v   \
                                                                in enumerate(inputIdxLst)       ]
                #write dataset to file
                # :-) startRow and startCol can have a valid value zero
                #GIOG-1282 quickfix instead of implementing intersect
                if (inEndRow < inStartRow) or (inEndCol < inStartCol):
                    continue
                dsOut[:,outStartRow:outEndRow,outStartCol:outEndCol] = inputBandArray[:,inStartRow:inEndRow,inStartCol:inEndCol]
                pass #debug line
    #move temp output file to actual output filename
    os.rename(tmpFile,outputFile)
    logger('Created: ' + outputFile)
    return 0,''

if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog='create a mosaic based on json config file')
    parser.add_argument('--cfgFile', type=str, 
                        help='the input configuration file')
    import glob
    '''
    #json file example
    {
        "input_files"          :
        "output_path"          :
        "output_roi_coords"    :
        "bands"                :
        "output_band_order"    :
        "output_pixel_step"    :
        "netcdf_time_ref_date" :
        "attributes"           :
        "productDate"          :
        "Optional"             : ''
        "input_band_order"     :
    }
    '''

    args = parser.parse_args()
    cd = os.path.dirname(os.path.realpath(__file__))
    input_path = os.path.dirname(args.cfgFile)
    nc_fileLst = glob.glob(os.path.join(input_path, '..','data', 'ut_cgl_BA300-NRT_202408010000_X*.nc'))
    nc_fileLst.extend(glob.glob(os.path.join('/vitodata/cgl_vol4/input/ancillary/missing_tiles/s3_bav3', '*missing_land_water_X07Y05*.nc')))
    nc_fileLst.extend(glob.glob(os.path.join('/vitodata/cgl_vol4/input/ancillary/missing_tiles/s3_bav3', '*missing_land_water_X01*.nc')))
    out_path   = os.path.join(cd, 'tmp')
    if not os.path.exists(out_path):
        os.makedirs(out_path)
    out_file   = os.path.join(out_path,'test_cgl_BA300-NRT_202408010000_GLOBE_S3_V3.1.1.nc')
    
    try:
        with open(args.cfgFile, encoding='utf-8') as fh:
            mosaicDict = json.load(fh)
        mosaicDict['input_files'] = nc_fileLst
        mosaicDict['output_file'] = out_file
        err, errMsg = createMosaicNetcdf(mosaicDict)
        if err:
            print(errMsg)
            raise Exception(errMsg)
        with nc.Dataset(out_file, 'r') as inFile:
            inFile.set_auto_maskandscale(False)
            ds = inFile.variables['day_of_burn']
            water = np.where(ds[:,0:5,3360:3370]==-2)
            if not len(water[0]):
                print ('water not found X01Y00')
            water = np.where(ds[:,15110:15120,3360:3370]==-2)
            if not len(water[0]):
                print ('water not found X01Y04')
            water = np.where(ds[:,15110:15120,26570:26580]==-2)
            if not len(water[0]):
                print ('water not found X07Y04')
        print('{} passed check for water'.format(out_file))
        os.remove(out_file)
    
    except:
        errMess  = str(sys.exc_info()[1])
        errLine  = str(sys.exc_info()[2].tb_lineno)
        exitMess = '[mosaic:' + errLine + '] failed:\n' + errMess
        print(1, exitMess)
        sys.exit(1)

    sys.exit(0)
