# -*- coding: utf-8 -*-
""" Copernicus Global Land

Script to delete products from the Object Store
Optional cleanup of the DB is possible (entries in the DB will get status OBSOLETE

usage: deleteFromOs.py [-h] --startDate --endDate 
                       --startDate {{params.startDate}} --endDate {{params.endDate}} --environ --sensor --product --period  --version --ancSource --updateDB 

options:
  -h, --help            show this help message and exit
  --startDate STARTDATE
                        the start date
  --endDate ENDDATE     the end date
  --environ {env,stag,prod,stag.custom,stag.it}
                        the environment prefix for the bucket, default taking
                        the env from the settings
  --sensor SENSOR       the sensor abbreviation
  --product PRODUCT     the product abbreviation
  --period PERIOD       the periodicity of the product as used in the bucket
                        name
  --version VERSION     the product version - can be empty for TOC
  --ancSource ANCSOURCE
                        the anciliary data source name
  --updateDB UPDATEDB   update DB with the products removed from waw3, only if
                        environ=env

e.g.:
deleteFromOs.py --startDate 20180625 --endDate 20180625 --environ env --sensor sen3 --product geo --period 1 --version v1.2.1 --updateDB True
deleteFromOs.py --startDate 20190620 --endDate 20190620 --sensor anc --product activefire --period 1 --version v1.2.1 --ancSource nrt --updateDB True
    
History
-------
========    =======  ======  ========  ========
date        version  author  jira ref  Comments
========    =======  ======  ========  ========
20250415    1.0.0    RVdG    GIOG-3605: make object storage cleanup DAG

"""
import argparse
import os
import glob
import json
import subprocess
import time, datetime
import traceback

from cgls_cpe.logging import log
from cgls_cpe.config.configuration import Configuration
from cgls_cpe.common import helper
from cgls_cpe.common.helper import _str2bool
from cgls_cpe.cpl.date_time import incrDate, today, listPeriod, checkLeapYear, daysInFebruary, getNrDaysInMonth
from cgls_cpe.storage import local_storage, remote_storage
from cgls_cpe.config.model import enums
from cgls_cpe.config import storage
from cgls_cpe.hspf import common_helpers
from cgls_cpe.config.model.ancillary_data import AncillaryData

from cgls_cpe.config.model.enums import AncillaryDataInputCategory,AncillaryDataSourceName

from cgls_cpe.db_api import DatabaseConnection, DB_product
from cgls_cpe.config.model.enums import DB_status, DB_product_type
from cgls_cpe.common.cdse_post_common import createPostCfgDictList
from numpy.core.numeric import False_


def run(cmd):
    """Run a command in the shell

    Parameters
    ----------
    cmd : str
        The command to be executed

    Returns
    -------
    subprocess.CompletedProcess
        The result of the command
    """
    try:
#        print('Executing command\n{}'.format(cmd))
        rslt = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
#        print('Execution output\n{}'.format(rslt.stdout.decode()))
        return rslt
    except:
        print('Error')
        traceback.print_exc()
        return None
    
def process(cfgDict):
    """Delete objects from Object Store

    Parameters
    ----------
    cfgDict : dict
        The configuration dictionary with info on products to be removed from waw3
    """
    config     = Configuration()
    settings   = config.get_settings()
    
    aws_s3_endpoint = settings.storage_end_point 
    aws_access_key_id = config.get_secrets().get_secret_key('AWS_ACCESS_KEY_ID')
    aws_secret_access_key = config.get_secrets().get_secret_key('AWS_SECRET_ACCESS_KEY')
    
    if cfgDict['environ'] == 'env':
        env   = os.environ['CGLS_CPE_ENV']
    else:
        env   = cfgDict['environ']
        if env != os.environ['CGLS_CPE_ENV'] and cfgDict['updateDB']:
            print(f"updateDB can only be used with {os.environ['CGLS_CPE_ENV']} - setting updateDB=False")
            cfgDict['updateDB'] = False
    sensor    = cfgDict['sensor']
    product   = cfgDict['product']
    periodicity = cfgDict['period']
    version   = cfgDict['version']
    startDate = cfgDict['startDate']
    endDate   = cfgDict['endDate']
    ancSource = cfgDict['ancSource']
    if sensor == 'anc' and not len(ancSource):
        print('Ancillary data source name is missing')
        exit(1)
    '''
    SatImages: buckets are having year-month as suffix
    AncillaryData: buckets are having year or year-month as suffix
    '''
    period = int(periodicity)
    # patch BA300 DB_product_type
    if product == 'BA300': 
        if period == 1:
            cfgDict['isNrt'] = True
        else:
            cfgDict['isNrt'] = False
        
    #force period for datelst for some products not having the full yyyy/mm/yyyymmdd structure
    if product in ['cbdk', 'tocr']:
        period = 30 #no 'date' subdirs
    if product in ['lsp']:
        period = 365
    dLst   = listPeriod(startDate, endDate, period=period, first_day_only=False)
    dateLst = [x for _,x in dLst]
    if ancSource == 'aer.extinction.monthly' or ancSource == 'surf.press.water.vapor.monthly':
        dateLst = [x for x,_ in dLst] #not likely to be removed - no DB entries ...
    if ancSource in ['cams.nrt.surface.ext']:
        print('Deleting cams.nrt.surface.ext not supported for now. Directory holding both nrt and ntc data. Not sure how to discriminate.')
        exit(1)
    if product in ['ndvi300', 'lsp']:
        dateLst = [x for x,_ in dLst]
    urlPrefix = 's3://'
    
    for date in dateLst:
        year  = date[0:4]
        month = date[4:6]
        dateSuffix = '-'.join([year, month])
        dateSuffix1 = '/'.join([year, month, date])
        if product in ['cbdk', 'tocr']:
            dateSuffix1 = '/'.join([year, month])
        if product in ['lsp']:
                dateSuffix = year
                dateSuffix1 = year
        if sensor == 'anc':
            if ancSource in ['aer.extinction.monthly', 'surf.press.water.vapor.monthly']:
                dateSuffix = year
                dateSuffix1 = year
            pattern =  '-'.join([env, sensor, product, ancSource, dateSuffix])
            pattern1 =  '/'.join(['', env, sensor, product, ancSource, dateSuffix1, ''])
        else:
            if len(version):
                pattern =  '-'.join([env, sensor, product, periodicity, version, dateSuffix])
                pattern1 =  '/'.join(['', env, sensor, product, periodicity, version, dateSuffix1, ''])
            else:
                pattern =  '-'.join([env, sensor, product, periodicity, dateSuffix])
                pattern1 =  '/'.join(['', env, sensor, product, periodicity, dateSuffix, ''])
        url_with_pattern = urlPrefix + pattern + pattern1
        #prod_present = remote_storage.list_objects_using_pattern(url_with_pattern)
        #removal via remote_storage removes individual objects
        #  ->  s3cmd can remove entire buckets
        #cmd = 's3cmd -c /home/gotenr/.s3cfg_waw3 ls {} '.format(url_with_pattern)
        cmd = f"export AWS_ACCESS_KEY_ID={aws_access_key_id}; export AWS_SECRET_ACCESS_KEY={aws_secret_access_key}; export AWS_S3_ENDPOINT={aws_s3_endpoint};"
        cmd += f's3cmd --host={aws_s3_endpoint} --host-bucket=https://{aws_s3_endpoint} ls {url_with_pattern} '
        rslt = run(cmd)
        if rslt.returncode != 0:
            print('Error s3cmd for ' + url_with_pattern )
            continue
        prodLst = rslt.stdout.decode().split('\n')
        print('{}: {} entries in ObjectStore'.format(date, len(prodLst)))
        if len(prodLst) and len(prodLst[0]):
            #cmd = 's3cmd -c /home/gotenr/.s3cfg_waw3 rb -r {}'.format(url_with_pattern)
            cmd = f"export AWS_ACCESS_KEY_ID={aws_access_key_id}; export AWS_SECRET_ACCESS_KEY={aws_secret_access_key}; export AWS_S3_ENDPOINT={aws_s3_endpoint};"
            cmd += f's3cmd --host={aws_s3_endpoint} --host-bucket=https://{aws_s3_endpoint} rb -r {url_with_pattern}'
            print(cmd)
            '''
            try:
                rslt = run(cmd)
                if rslt.returncode != 0:
                    print('Error ' + cmd )
            except:
                print('Error removing products from ObjectStore {}'.format(url_with_pattern))
            '''
    if cfgDict['updateDB']:
        from updateDb import process
        process(cfgDict)
    return

if __name__ == '__main__':
    parser = argparse.ArgumentParser(prog='delete objects from cdso object store')
    parser.add_argument('--startDate', type=str, 
                        help='the start date')
    parser.add_argument('--endDate', type=str, 
                        help='the end date')
    parser.add_argument('--environ', type=str, default='env',
                        choices=['env', 'stag', 'prod', 'stag.custom', 'stag.it'],
                        help='the environment prefix for the bucket, default taking the env from the settings')
    parser.add_argument('--sensor', type=str,
                        default='sen3',
                        help='the sensor abbreviation')
    parser.add_argument('--product', type=str,
                        help='the product abbreviation ')
    parser.add_argument('--period', type=str, default='1',
                        help='the periodicity of the product as used in the bucket name')
    parser.add_argument('--version', type=str,
                        default='',
                        help='the product version - can be empty for TOC')
    parser.add_argument('--ancSource', type=str, default='',
                        help='the anciliary data source name ')
    parser.add_argument('--updateDB', type=_str2bool, default=False,
                        help='update DB with the products removed from waw3, only if environ=env')
    args = parser.parse_args()

    #dunder merge https://www.dynaconf.com/merging/
    
    os.environ['CGLS_CPE_LOGGING_CONFIG__handlers__default__level'] = 'DEBUG'
    os.environ['CGLS_CPE_LOGGING_CONFIG__loggers__cgls__level'] = 'DEBUG'
    
    '''
    add products from /vitodata to ObjectStore
    add to stag - prod depending on the setting of the following env
    CGLS_CPE_ENV CGLS_CPE_VAULT_ROLE_ID CGLS_CPE_VAULT_SECR_ID
    
    execute the following command to install the credentials:
    s3cmd --configure
        Access Key: xxx - refer to Vault
        Secret Key: xxx 
        Default Region: US
        S3 Endpoint:  s3.waw3-2.cloudferro.com  
        DNS-style bucket+hostname:port template for accessing a bucket: s3.waw3-2.cloudferro.com
        Encryption password:
        Path to GPG program: /usr/bin/gpg
        Use HTTPS protocol: Yes
        HTTP Proxy server name:
        HTTP Proxy server port: 0
    cp ~/.s3cfg ~/.s3cfg_aws

    '''
    process(args.__dict__)
    
