'''
Created on Nov 5, 2023

@author: demunckd
'''
import re
import json
from glob import glob

from cgls_cpe.config.model.context import Context
from cgls_cpe.config.configuration import Configuration

from cgls_cpe.storage.storage_helper import S3PREFIX,VSIS3PREFIX
from cgls_cpe.storage.implementation.s3_api import S3Api
from cgls_cpe.storage.implementation.s3_api import get_s3_url,is_s3_or_vsis
from cgls_cpe.common import helper
from cgls_cpe.config.model.data_object import DataObject
from numba.core.types import none
try: 
    from smart_open import (
        open as s_open,  # this will make openening a file local/s3 transparant
    )

except:
    print("SMART OPEN IMPORT FAILED -")
import os

NO_CUSTOM_STORAGE_ENV = ''
custom_storage_env_id = NO_CUSTOM_STORAGE_ENV
s3_api_default = None
s3_api_custom_storage_env = {}


from cgls_cpe.logging import log

logger = log.logger()


SUFFIX_ENV_ACCESS_KEY_ID = '_ACCESS_KEY_ID'
SUFFIX_ENV_SECRET_ACCESS_KEY = '_SECRET_ACCESS_KEY'
SUFFIX_ENV_ENDPOINT = '_ENDPOINT'


#set None if no custom env but use default keys
def set_custom_env(custom_id:str):
    global custom_storage_env_id
    if custom_id is None:
        custom_storage_env_id = NO_CUSTOM_STORAGE_ENV
    else:
        custom_storage_env_id = custom_id

@helper.url_s3_safe
def get_key_from_url(url):
    if url.startswith(S3PREFIX):
        no_prefix = url[len(S3PREFIX):]
        key = no_prefix[(no_prefix.index('/')+1):]
        return key


def check_environment(context):
    if context == Configuration().get_storage().get_reference_context():
        return
    if context.is_production() != Configuration().is_production():
        raise Exception("Mixing up environments")


def is_valid_bucket_name(bucket_name):
    """
    Validates if a string is a valid bucket name according to AWS S3 naming rules.
    """
    # Bucket name must be between 3 and 63 characters
    if len(bucket_name) < 3 or len(bucket_name) > 63:
        return False

    # Bucket name can only contain lowercase letters, numbers, hyphens, and periods
    if not re.match("^[a-z0-9.-]+$", bucket_name):
        return False

    # Bucket name cannot start or end with a hyphen or period
    if bucket_name.startswith("-") or bucket_name.startswith("."):
        return False
    if bucket_name.endswith("-") or bucket_name.endswith("."):
        return False

    # Bucket name cannot have consecutive periods or hyphens
    if ".." in bucket_name or "--" in bucket_name:
        return False

    return True

#### SAT_IMAGE DOWNLOAD/STORE

def get_full_URL(image:DataObject,context:Context=None, use_vsis3_prefix=False):
    if context is None:
        context = Configuration().get_storage().get_default_context()
    
    container = Configuration().get_storage().get_remote_container_for_image(image,context)
    key = Configuration().get_storage().get_remote_key_for_image(image,context)    
    return get_s3_url(container,key, use_vsis3_prefix=use_vsis3_prefix)

def get_location(image:DataObject,context:Context=None):
    full_url = get_full_URL(image,context)
    return full_url.rsplit('/', 1)[0]


def store_remotely(local_file:str, image:DataObject,context:Context=None):
    container,context = prepare_storage(image, context)
    key = Configuration().get_storage().get_remote_key_for_image(image,context)
    
    __get_s3_api().upload_single_file(local_file, container, key)


def prepare_storage(image:DataObject,context:Context=None):
    if context is None:
        context = Configuration().get_storage().get_default_context()
    else:
        check_environment(context)
    
    container = Configuration().get_storage().get_remote_container_for_image(image,context)
    __create_container_with_correct_permissions_if_requried(container)
    return container,context


def remove_remotely(image:DataObject,context:Context=None):
    if context is None:
        context = Configuration().get_storage().get_default_context()
    else:
        check_environment(context)
    
    container = Configuration().get_storage().get_remote_container_for_image(image,context)
    key = Configuration().get_storage().get_remote_key_for_image(image,context)    
    __get_s3_api().remove_object(container, key)


def download_remote(image:DataObject,dest_file,context:Context=None):
    if context is None:
        context = Configuration().get_storage().get_default_context()
    else:
        check_environment(context)

    container = Configuration().get_storage().get_remote_container_for_image(image,context)
    key = Configuration().get_storage().get_remote_key_for_image(image,context)    
    __get_s3_api().download_single_file(container, key, dest_file)


def list_remote(image: DataObject, context: Context = None, use_vsis3_prefix=False):
    if context is None:
        context = Configuration().get_storage().get_default_context()
    else:
        check_environment(context)

    container = Configuration().get_storage().get_remote_container_for_image(image,context)
    key = Configuration().get_storage().get_remote_key_for_image(image,context)
    res = __get_s3_api().list_by_prefix(container,key,return_s3_url_as_keys=False)[1]
    if use_vsis3_prefix:
        res = __replace_s3_prefix(res)
    return res


### GENERIC URL COMMANDS

def is_posix_storage(url):
    if is_s3_or_vsis(url):
        return False
    return True

def load_json_from_url(url):
    with open_smart(url) as f:
        data = json.load(f)
    return data


#works for POSIX and REMOTE
def object_exists(url):
    if is_s3_or_vsis(url):
        container, key = __get_s3_api().get_bucket_from_url(url)
        return __get_s3_api().check_if_key_exists(container, key)
    else:
        return os.path.exists(url)

# returns urls with given prefix
@helper.url_s3_safe
def list_objects(url, use_vsis3_prefix=False):
    container, key = __get_s3_api().get_bucket_from_url(url)
    res:list[str] = __get_s3_api().list_by_prefix(container,key,return_s3_url_as_keys=True)[1]
    if use_vsis3_prefix:
        res = __replace_s3_prefix(res)
    return res


# url_with_pattern is s3:// style
# or if alternative_s3_prefix is set then use this one
@helper.url_s3_safe
def list_objects_using_pattern( url_with_pattern, use_vsis3_prefix=False ):
    
    if is_s3_or_vsis(url_with_pattern):
        container, key = __get_s3_api().get_bucket_from_url(url_with_pattern)
        prefix = key
        if  '*' in key:
            prefix = key.split('*')[0]   #remove possible leading slash
        res = __get_s3_api().list_by_pattern(container, prefix, key)
        if use_vsis3_prefix:
            res = __replace_s3_prefix(res)
        return res
    else:
        return glob(url_with_pattern)
        

def download_remote_to_local_folder(remotekeys, local_path, overwrite=True, options=''):
    options='--force'
    if not overwrite :
        options = '--skip-existing'
        
    return __get_s3_api().download_from_s3(remotekeys, local_path,options)

def download_list_of_keys_to_local_folder(remotekeys, local_path, overwrite=True, max_workers=None):
    return __get_s3_api().download_list_of_keys_to_local_folder(
        remotekeys, local_path, overwrite, max_workers
    )

@helper.url_s3_safe
def upload_local_to_remote(local_files, s3_path, rename_during_transfer=True, options=''):
    return __get_s3_api().upload_to_s3(local_files, s3_path, rename_during_transfer, options) 

def sync_local_folder_to_remote(local_folder, url,options='--exact-timestamps'):
    container, _ = __get_s3_api().get_bucket_from_url(url)
    __create_container_with_correct_permissions_if_requried(container)    
    return __get_s3_api().sync_folders_via_aws(local_folder,url,options)

def sync_remote_to_local_folder(local_folder, url,options='--exact-timestamps'):
    return __get_s3_api().sync_folders_via_aws(url, local_folder,options)

@helper.url_s3_safe
def upload_single_file_to_remote_path(local_path, remote_path):
    container, key = __get_s3_api().get_bucket_from_url(remote_path)
    __create_container_with_correct_permissions_if_requried(container)    
    __get_s3_api().upload_single_file(local_path, container,key)
    return 0

@helper.url_s3_safe
def download_single_file_to_local_path(remote_url, dest_file):
    bucket, key = __get_s3_api().get_bucket_from_url(remote_url)
    __get_s3_api().download_single_file(bucket, key, dest_file) 
    return 0
    
@helper.url_s3_safe
def get_contents_as_string(url):
    return __get_s3_api().get_contents_as_string(url)

@helper.url_s3_safe
def put_contents_as_string_to_url(url,input_str):
    return __get_s3_api().put_contents_as_string_to_url(url, input_str)

@helper.url_s3_safe
def copy_or_move_object(src_url, tgt_url, is_move=False):
    return __get_s3_api().copy_or_move_object(src_url, tgt_url, is_move)

@helper.url_s3_safe
def remove_object(url):
    bucket, key = __get_s3_api().get_bucket_from_url(url)
    __get_s3_api().remove_object(bucket, key)

@helper.url_s3_safe
def remove_objects(url):
    existing_objects = list_objects(url)
    for obj  in existing_objects :
        remove_object(obj)


def open_smart(url, mode='r'):
    try:
        if url.startswith(VSIS3PREFIX) or url.startswith(S3PREFIX):
            return __get_s3_api().open_smart(url,mode)    
        else:
            return(s_open(url,mode=mode, transport_params={}))
    except:
        logger.info("skipping smart open because of issue - probably import failed")
        return open(url,mode=mode)
        

def export_credentials_to_env():
    __get_s3_api().export_credentials_to_env()

def mount_s3fs(bucket_name, mount_point):
    helper.mkdir_p(mount_point)
    
    cmd = 's3fs  ' + bucket_name + ' ' + mount_point + ' -o url=https://' + Configuration().get_storage().get_object_storage_endpoint() + ' -o use_path_request_style '
    logger.info("Going to execute mount cmd: " + cmd)
    helper.exec_with_env( cmd, __get_s3_api().get_env_for_object_storage())
    logger.info("Mounting finished")

def unmount_s3fs(mount_point):
    cmd = 'fusermount -u ' + mount_point
    logger.info("Going to execute umount cmd: " + cmd)
    helper.exec_with_env(cmd)
    logger.info("Unmounting finished")

# trigger

def get_process_status_container(context:Context=None):
    if context is None:
        context = Configuration().get_storage().get_default_context()
    else:
        check_environment(context)
    
    container =  '%s-processing-status' % context.get_value()
    __create_container_with_correct_permissions_if_requried(container)
    return container


def get_temp_container(context:Context=None):
    if context is None:
        context = Configuration().get_storage().get_default_context()
    else:
        check_environment(context)
    
    container =  '%s-temp' % context.get_value()
    __create_container_with_correct_permissions_if_requried(container)
    return container



# INTERNAL FUNCTIONS, should normally NOT be called by external

def reset_auth():
    global s3_api_default,s3_api_eodata
    s3_api_default = None
    s3_api_custom_storage_env.clear()

def __get_s3_api():
    
    global custom_storage_env_id
    global s3_api_custom_storage_env
    global s3_api_default
    
    store = Configuration().get_storage()
    if custom_storage_env_id != NO_CUSTOM_STORAGE_ENV:
        if custom_storage_env_id in s3_api_custom_storage_env:
            logger.debug("Using custom storage env id: " + custom_storage_env_id)
            return s3_api_custom_storage_env[custom_storage_env_id]
        else:
            key_access = custom_storage_env_id + SUFFIX_ENV_ACCESS_KEY_ID
            if  key_access in os.environ:
                value_access_id = os.getenv(key_access)
                key_secret =  custom_storage_env_id + SUFFIX_ENV_SECRET_ACCESS_KEY
                if  key_secret in os.environ:
                    value_secret_key = os.getenv(key_secret)
                    value_endpoint = store.get_object_storage_endpoint()
                    key_endpoint =  custom_storage_env_id + SUFFIX_ENV_ENDPOINT
                    if key_endpoint in os.environ:
                        value_endpoint = os.getenv(key_endpoint)
                    s3_api_custom_storage_env[custom_storage_env_id] = S3Api(value_endpoint, value_access_id,value_secret_key)
                    logger.debug("Created custom storage with env id" + custom_storage_env_id + " with values: %s %s %s", value_endpoint, value_access_id,value_secret_key)
                    return s3_api_custom_storage_env[custom_storage_env_id]
                           

    if s3_api_default is None:
            s3_api_default = S3Api(store.get_object_storage_endpoint(), store.get_object_storage_acces_id(), store.get_object_storage_secret_acces_key())
    return s3_api_default
    

def __create_container_with_correct_permissions_if_requried(container):
    bucket_exists = __get_s3_api().check_if_bucket_exists(container)
    if not bucket_exists:
        __get_s3_api().create_bucket(container)
        custom_policy = Configuration().get_storage().get_s3_policy_for_container(container)
        if custom_policy is not None:
            __get_s3_api().set_policy_on_bucket(container, custom_policy) 

def __replace_s3_prefix(s3_url_strings):
    if not isinstance(s3_url_strings, list):
        s3_url_strings = [s3_url_strings]

    s3_url_strings = [
        s3_url.replace(S3PREFIX, VSIS3PREFIX, 1) for s3_url in s3_url_strings
    ]

    return s3_url_strings
