import argparse
from datetime import datetime

from cgls_cpe.config.configuration import Configuration
from cgls_cpe.config.model.enums import DB_status, FileTypeExtension
from cgls_cpe.config.model.version import Version
from cgls_cpe.cpl.netcdf import compare_nc_files
from cgls_cpe.db_api import DatabaseConnection
from cgls_cpe.logging import log
from cgls_cpe.storage import remote_storage
from email.mime import image

logger = log.logger()

parser = argparse.ArgumentParser(prog="test_results.py")

### required arguments ###
parser.add_argument(
    "-pti",
    "--product_type_id",
    required=True,
    help="Product type id",
)

parser.add_argument(
    "-d",
    "--date",
    required=False,
    default="20191011",
    help="Target date to search on the database for validation",
)
parser.add_argument(
    "-v", "--version", required=True, help="Target version including tc, e.g. 2.3.tc4"
)

parser.add_argument(
    "-c", "--count", default=1, required=False, help="Number of files to compare"
)

parser.add_argument(
    "-r",
    "--ref_version",
    required=False,
    help="Reference version to compare against, if not provided the last before test candidate will be used e.g.for 2.3.tc4 the reference will be 2.3.3",
)

parser.add_argument(
    "--options",
    required=False,
    default=["SKIP_METADATA", "SKIP_GEOLOCATION"],
    nargs="+",
    help="Options for gdal comparison, usually you prefer to skip metadata and geolocation",
)

args = parser.parse_args()
if not args.ref_version:
    tc = Version.from_string(args.version)
    ref_v = Version(tc.major, tc.minor, tc.update - 1)
    args.ref_version = ref_v.get_value()

remote_storage.export_credentials_to_env()
db_settings = Configuration().get_database()
db = DatabaseConnection()


def validate_results():

    date = datetime.strptime(args.date, "%Y%m%d")
    start_datetime = datetime(date.year, date.month, date.day, 0, 0, 0)
    end_datetime = datetime(date.year, date.month, date.day, 23, 59, 59)

    res = db.get_products(
        product_type=args.product_type_id,
        status_lst=DB_status.PROCESSED.value,
        startdate=start_datetime,
        enddate=end_datetime,
        version=args.version,
    )

    if not res:
        logger.error(
            f"No products found for date {args.date} and version {args.version}"
        )
        raise Exception(
            f"No products found for date {args.date} and version {args.version}"
        )
    i = 1
    logger.info("Going to compare %s files as requested", args.count)    
    for file in res:
        location = file.location
        if not location.endswith(FileTypeExtension.NC.value):
            image_files = remote_storage.list_objects(location);
            for image in image_files:
                if ( image.endswith(FileTypeExtension.NC.value)):
                    location = image
                    break
        ref_key = location.replace(args.version, args.ref_version)
        logger.info("Comparing test product %s with ref %s and options %s", location, ref_key, args.options )
        if not compare_nc_files(location, ref_key, args.options):
            raise Exception(f"Validation failed for product {location}")
        else:
            logger.info("Compare OK")

        if i >= int(args.count):
            break
        i +=1
        

    logger.info("Validation passed for all products")


if __name__ == "__main__":
    validate_results()
